#!/bin/bash

# Must have sudo permissions to run

# SETUP ------------------------------------------

# This script inherits the following variables from configure-git-fusion.sh
# If running this script by itself, set values for these required variables

# P4 variables:
# P4PORT=
# P4TICKETSPATH=
# P4TRUSTPATH=

# GCONN_CONFIG path:
# GCONNCONFIGPATH=

# Unix group used for authentication:
# RUN_GROUP=

# Interactive mode (defaults to true)
INTERACTIVE=${INTERACTIVE:-true}

if [ "$P4PORT" == '' ]; then
    echo 'P4PORT is undefined'
    exit 1
fi
if [ "$P4TICKETSPATH" == '' ]; then
    echo 'P4TICKETSPATH is undefined'
    exit 1
fi
if [ "$P4TRUSTPATH" == '' ]; then
    echo 'P4TRUSTPATH is undefined'
    exit 1
fi
if [ "$GCONNCONFIGPATH" == '' ]; then
    echo 'GCONNCONFIGPATH is undefined'
    exit 1
fi
if [ "$RUN_GROUP" == '' ]; then
    echo 'RUN_GROUP is undefined'
    exit 1
fi
if [ "$GCONN_HOSTNAME" == '' ]; then
    echo "GCONN_HOSTNAME is undefined"
    exit 1
fi

# HELPER FUNCTIONS ------------------------------

debug()
{
    $DEBUG || return 0
    $MONOCHROME || echo -e "\033[33m"$1"\033[0m" >&2
    $MONOCHROME && echo -e "$1" >&2 || true
}

die()
{
    error "FATAL: $@" >&2
    exit 1
}

error()
{
    $MONOCHROME || echo -e "\033[31m"$1"\033[0m" >&2
    $MONOCHROME && echo -e "$1" >&2 || true
}

highlight()
{
    $MONOCHROME || echo -e "\033[32m"$1"\033[0m"
    $MONOCHROME && echo -e "$1" || true
}


# Validates a yes/no prompt
# Usage: validate_yesno [yes/no>
validate_yesno()
{
    local CHOICE="$1"
    if [ -z "$CHOICE" ] || [ "$CHOICE" != "yes" -a "$CHOICE" != "no" ]; then
        echo 'Please specify either yes or no'
        return 1
    fi
    return 0
}

# Prompt the user for information by showing a prompt string. Optionally
# calls a validation function to check if the response is OK.
#
# promptfor <VAR> <prompt> <default> [<validationfunc>]
promptfor()
{
    local var="$1"
    local prompt="$2"
    local default="$3"
    local check_func=true

    [ -n "$4" ] && check_func=$4

    if [ "$default" = " " ]; then
        default=""
    fi

    while true; do
        local pw=""
        if [ -n "$default" ]; then
            read -e -p "$prompt [$default]: " pw
            if [ ! -n "$pw" ]; then
                pw=$default
            fi
        else
            read -e -p "$prompt: " pw
        fi
        stty echo -echonl || true
        if $check_func "$pw"; then
            eval "$var=\"$pw\""
            break;
        fi
    done
    true
}
# ASSESS ENVIRONMENT -----------------------------
if [ -e '/etc/lsb-release' ] && [ ! -e '/etc/redhat-release' ] && [ ! -e '/etc/SuSE-release' ]; then
    UBUNTU_VERSION="$(cat /etc/lsb-release | grep DISTRIB_RELEASE | cut -d '=' -f 2 )"
elif [ -e '/etc/redhat-release' ]; then
    CENTOS_VERSION="$(grep -oE '[0-9]+\.[0-9]+' /etc/redhat-release)"
elif [ -e '/etc/SuSE-release' ]; then
    SUSE_VERSION="$(sed -n 's/VERSION = //p' /etc/SuSE-release)"
else
    echo 'Could not determine your OS version. Aborting https installation.'
    exit 0
fi

if [ "$UBUNTU_VERSION" == '12.04' ]; then
    error "You Ubuntu version is [$UBUNTU_VERSION]. HTTPS authentication for"
    error "GitConnector can only be configured on Ubuntu 14.04 and later."
    exit 1
fi

# Check if Apache is installed. If yes, warn users
APACHE_INSTALLED=false
APACHE_DIR_EXISTS=false
if [ -n "$UBUNTU_VERSION" ]; then
    APACHE_PKG=$(dpkg -l 'apache2' 2>&1 | grep 'apache2' || true)
    if [ -n "$APACHE_PKG" ]; then
        STATUS=$(echo "$APACHE_PKG" | awk '{print $1}');
        if [ "$STATUS" = 'ii' ]; then
            APACHE_INSTALLED=true
        fi
    fi
elif [ -n "$CENTOS_VERSION" ]; then
    APACHE_PKG=$(rpm -qa 'httpd' 2>&1 | grep 'httpd')
    APACHE_INSTALLED=true
elif [ -n "$SUSE_VERSION" ]; then
    APACHE_PKG=$(rpm -qa 'apache2' 2>&1 | grep 'apache2')
    if [ -n "$APACHE_PKG" ]; then
        APACHE_INSTALLED=true
    fi
fi

if [ -e '/etc/apache2' ] || [ -e '/etc/httpd' ]; then
    APACHE_DIR_EXISTS=true
fi
if $APACHE_INSTALLED || $APACHE_DIR_EXISTS; then
    highlight 'It looks like Apache components are already present on this machine.'
		highlight 'GitConnector HTTPS automated configuration will significantly modify the
Apache configuration, and will disable existing sites.'
    if $INTERACTIVE; then
        promptfor CONFIGURE_HTTP 'Do you wish to continue?' 'yes' validate_yesno
        if [ "$CONFIGURE_HTTP" != 'yes' ]; then
            error 'Stopping.'
            exit 1
        fi
    elif $FORCE_HTTPS; then
        highlight "$SCRIPTPATH/$SCRIPT is called with
--forcehttps flag. Continuing HTTPS configuration."
    else
        error "Run $SCRIPTPATH/$SCRIPT in interactive mode or"
        error "with --forcehttps flag to continue HTTPS configuration."
        exit 1
    fi
else
    highlight 'This script will now configure HTTPS authentication for Helix GitConnector using Apache.'
    if [ -n "$UBUNTU_VERSION" ]; then
        highlight 'It will download and install Apache 2 and enable modules "ssl", "rewrite" and "cgi".'
    fi
    highlight 'If you have existing Apache configuration, it may be overwritten.
It is not recommended that you use this Apache server for any other purpose than for use with Helix GitConnector.'
fi

# CONFIG -----------------------------------------

# Install Apache and modules
if [ -n "$UBUNTU_VERSION" ]; then
    APACHE_INSTALL=$(apt-get install -y apache2 2>&1)
    ret=$?
    APACHE_PKG=$(dpkg -l 'apache2' 2>&1 | grep 'apache2' || true)
elif [ -n "$CENTOS_VERSION" ]; then
    APACHE_INSTALL=$(yum install -y httpd mod_ssl 2>&1)
    ret=$?
    APACHE_PKG=$(rpm -qa 'httpd' 2>&1 | grep 'httpd')
elif [ -n "$SUSE_VERSION" ]; then
    APACHE_INSTALL=$(zypper --non-interactive install apache2 2>&1)
    ret=$?
    APACHE_PKG=$(rpm -qa 'apache2' 2>&1 | grep 'apache2')
fi

# Check that install was successfull
if [ -n "$APACHE_PKG" ] && [ $ret -eq 0 ]; then
    if [ -n "$UBUNTU_VERSION" ]; then
        STATUS=$(echo "$APACHE_PKG" | awk '{print $1}');
        if [ "$STATUS" = 'ii' ]; then
            highlight 'Apache was installed successfully'
        fi
    elif [ -n "$CENTOS_VERSION" ]; then
        highlight 'Apache was installed successfully'
    elif [ -n "$SUSE_VERSION" ]; then
        highlight 'SUSE Apache was installed successfully'
    fi
else
    error 'Apache or one of its components could not be installed!'
    error 'Here is the detailed output:'
    error "$APACHE_INSTALL"
    exit 1
fi

if [ -n "$CENTOS_VERSION" ]; then
    # Disable SELinux for the current session
    setenforce 0 > /dev/null 2>&1 || true
    systemctl disable firewalld > /dev/null 2>&1 || true
    systemctl stop firewalld > /dev/null 2>&1 || true
fi


# Set required variables which are platform-dependent
if [ -n "$UBUNTU_VERSION" ]; then
    APACHE_SERVICE_NAME='apache2'
    APACHE_GLOBAL_CONF='/etc/apache2/apache2.conf'
    APACHE_GCONN_CONFIG_FILE='/etc/apache2/sites-available/gconn.conf'
    APACHE_SYS_USER='www-data'
    SSL_DIR='/etc/apache2/ssl'
elif [ -n "$CENTOS_VERSION" ]; then
    APACHE_SERVICE_NAME='httpd'
    APACHE_GLOBAL_CONF='/etc/httpd/conf/httpd.conf'
    APACHE_GCONN_CONFIG_FILE='/etc/httpd/conf.d/gconn.conf'
    APACHE_SYS_USER='apache'
    SSL_DIR='/etc/httpd/ssl'
elif [ -n "$SUSE_VERSION" ]; then
    APACHE_SERVICE_NAME='apache2'
    APACHE_GLOBAL_CONF='/etc/apache2/httpd.conf'
    APACHE_GCONN_CONFIG_FILE='/etc/apache2/vhosts.d/gconn.conf'
    APACHE_SYS_USER='wwwrun'
    SSL_DIR='/etc/apache2/ssl'
fi

# Stop Apache for duration of configuration
if [ -n "$SUSE_VERSION" ]; then
    /sbin/service $APACHE_SERVICE_NAME start > /dev/null
    /sbin/service $APACHE_SERVICE_NAME stop > /dev/null
else
    service $APACHE_SERVICE_NAME stop > /dev/null
fi

if [ -n "$UBUNTU_VERSION" ]; then
    # Disable all existing sites
    a2dissite '*' > /dev/null
    # Enable required modules
    APACHE_MOD_ENABLE=$(a2enmod ssl cgi rewrite 2>&1)
    if [ $? -ne 0 ]; then
        error 'Required Apache modules could not be enabled!'
        echo 'Here is the detailed output:'
        echo "$APACHE_MOD_ENABLE"
        exit 1
    else
        debug 'Enabled the required Apache modules'
    fi
elif [ -n "$CENTOS_VERSION" ]; then
    if [ -e /etc/httpd/conf.d/ssl.conf ]; then
        debug 'Removing default configuration file /etc/httpd/conf.d/ssl.conf'
        rm -rf /etc/httpd/conf.d/ssl.conf
        if [ $? -ne 0 ]; then
            error 'Could not delete default configuration file /etc/httpd/conf.d/ssl.conf'
            exit 1
        fi
    fi
elif [ -n "$SUSE_VERSION" ]; then
    # Enable required modules
    for mod in ssl cgi env rewrite
    do
		APACHE_MOD_ENABLE=$(a2enmod $mod 2>&1)
		if [ $? -ne 0 ]; then
			error 'Required Apache modules could not be enabled!'
			echo 'Here is the detailed output:'
			echo "$APACHE_MOD_ENABLE"
			exit 1
		else
			debug 'Enabled the required Apache module $mod'
		fi
    done
    APACHE_FLAG_ENABLE=$(a2enflag SSL 2>&1)
    if [ $? -ne 0 ]; then
        error 'Required Apache SSL mod could not be activated (a2enflag SSL)'
        echo 'Here is the detailed output:'
        echo "$APACHE_FLAG_ENABLE"
        exit 1
    else
        debug 'Activated Apache SSL support.'
    fi
fi

# Append ServerName to apache main configuration file
if ! grep -q "ServerName $GCONN_HOSTNAME" $APACHE_GLOBAL_CONF; then
    echo "ServerName $GCONN_HOSTNAME" >> $APACHE_GLOBAL_CONF
fi

# Comment out 'AddDefaultCharset' section in Apache config
# it breaks git clients < 2.1 when cloning special (@info, @help, @list, etc) commands
if [ -n "$CENTOS_VERSION" ]; then
    sed -i -e "s|AddDefaultCharset[^/]*|#AddDefaultCharset[^/]*|" $APACHE_GLOBAL_CONF
fi

# Check if GCONN Apache config already exists
if [ -e "$APACHE_GCONN_CONFIG_FILE" ]; then
    highlight "There is an existing Apache virtual host for GitConnector"
    highlight "This configuration will remove all existing files and create new ones"
    if $INTERACTIVE; then
        promptfor REMOVE_EXISTING_SITE 'Do you wish to continue?' 'yes' validate_yesno
        if [ "$REMOVE_EXISTING_SITE" != 'yes' ]; then
            error 'Stopping.'
            exit 1
        fi
    elif $FORCE_HTTPS; then
        highlight "$SCRIPTPATH/$SCRIPT is called with
--forcehttps flag. Continuing HTTPS configuration."
    else
        error "Run $SCRIPTPATH/$SCRIPT in interactive mode or"
        error "with --forcehttps flag to continue HTTPS configuration."
        exit 1
    fi
    highlight "Removing $APACHE_GCONN_CONFIG_FILE"
    if ! rm -f "$APACHE_GCONN_CONFIG_FILE"; then
        error "Failed to remove $APACHE_GCONN_CONFIG_FILE"
        exit 1
    fi
fi

# Create site for GConn
debug "Writing Apache virtual host for GitConnector"
if [ -n "$UBUNTU_VERSION" ]; then
    cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
<IfModule mod_ssl.c>
GCONN_SITE
elif [ -n "$CENTOS_VERSION" ]; then
    cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
LoadModule ssl_module modules/mod_ssl.so
Listen 443
GCONN_SITE
elif [ -n "$SUSE_VERSION" ]; then
    cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
<IfDefine SSL>
<IfDefine !NOSSL>
GCONN_SITE
fi
cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
<VirtualHost *:443>
    SSLEngine on
    SSLCertificateFile $SSL_DIR/apache.crt
    SSLCertificateKeyFile $SSL_DIR/apache.key
    SetEnv GCONN_CONFIG $GCONNCONFIGPATH
    ScriptAliasMatch \\
      "(?x)^/(.*/(HEAD | \\
          info/refs | \\
          info/lfs/locks | \\
          info/lfs/locks/verify | \\
          info/lfs/locks/.*/unlock | \\
          objects/(batch | \\
          [0-9a-f]{2}/[0-9a-f]{2}/[0-9a-f]{60} | \\
          [0-9a-f]{64} | \\
          pack/pack-[0-9a-f]{40}\.(pack|idx)) | \\
          git-(upload|receive)-pack) | \\
          mirrorhooks-list | \\
          mirrorhooks-add | \\
          mirrorhooks-remove | \\
          mirrorhooks)$" \\
    /usr/bin/gconn/\$1

    RewriteEngine On
    RewriteCond %{HTTP:Authorization} ^(.*)
    RewriteRule .* - [e=HTTP_AUTHORIZATION:%1]
    <Location />
        AuthType None
        Require all granted
GCONN_SITE
if [[ "$CENTOS_VERSION" == '6.'* ]]; then
    cat >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
        Satisfy any
GCONN_SITE
fi
if [[ "$SUSE_VERSION" == '11' ]]; then
    cat >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
        Allow from all
        Satisfy any
GCONN_SITE
fi
cat >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
    </Location>
GCONN_SITE

if [ -n "$SUSE_VERSION" ]; then
    cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
    <Directory "/usr/bin/">
       Options +ExecCGI +FollowSymLinks
       SetHandler cgi-script
    </Directory>
GCONN_SITE
fi
cat >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
</VirtualHost>
GCONN_SITE
if [ -n "$SUSE_VERSION" ]; then
    cat  >>"$APACHE_GCONN_CONFIG_FILE" <<GCONN_SITE
</IfDefine>
</IfDefine>
GCONN_SITE
fi

if [ ! -f "$APACHE_GCONN_CONFIG_FILE" ]; then
    error "Failed to write Apache virtual host"
    exit 1
fi

if [ -n "$UBUNTU_VERSION" ]; then
    # Enable gconn site
    debug "Enabling Apache virtual host for GitConnector"
    APACHE_GCONN_ENABLE=$(a2ensite gconn 2>&1)
    if [ $? -ne 0 ]; then
        error 'Could not enable Apache virtual host for GitConnector!'
        echo 'Here is the detailed output:'
        echo "$APACHE_GCONN_ENABLE"
        exit 1
    else
        debug 'Enabled Apache virtual host for GitConnector'
    fi
fi
# Add default Apache system user to "$RUN_GROUP"
debug "Adding [$APACHE_SYS_USER] system user to system group [$RUN_GROUP]"
apache_groups=$(groups $APACHE_SYS_USER)
if [[ "$SUSE_VERSION" == '11' ]]; then
	if ! id -Gn "$APACHE_SYS_USER" | grep &>/dev/null "$RUN_GROUP"; then
	ALL_GROUPS=$(groups $APACHE_SYS_USER | sed "s/.*: //;s/ /,/g"),$RUN_GROUP
	# SUSE 11 usemod requires all group names to be listed - there is no append 
	usermod -G "$ALL_GROUPS" "$APACHE_SYS_USER"
fi
else
    usermod -a -G "$RUN_GROUP" $APACHE_SYS_USER
fi
apache_groups=$(groups $APACHE_SYS_USER)
if id -Gn "$APACHE_SYS_USER" | grep &>/dev/null "$RUN_GROUP"; then
    debug "User [$APACHE_SYS_USER] belongs to group [$RUN_GROUP]"
else
    die "User [$APACHE_SYS_USER] does not belong to group [$RUN_GROUP]"
fi

# Check if SSL certificates exists
if [ -e "$SSL_DIR" ] ||\
    [ -e "$SSL_DIR/apache.key" ] || \
    [ -e "$SSL_DIR/apache.crt" ]; then
    highlight "There existing SSL certificates in $SSL_DIR"
    highlight "They will be removed and new SSL certificates will be generated"
    if $INTERACTIVE; then
        promptfor REMOVE_EXISTING_SSL_CERTS 'Do you wish to continue?' 'yes' validate_yesno
        if [ "$REMOVE_EXISTING_SSL_CERTS" != 'yes' ]; then
           error 'Stopping.'
           exit 1
        fi
        highlight "Removing $SSL_DIR"
        if ! rm -rf "$SSL_DIR"; then
            error "Failed to remove $SSL_DIR"
            exit 1
        fi
    elif $FORCE_HTTPS; then
        highlight "$SCRIPTPATH/$SCRIPT is called with
--forcehttps flag. Continuing HTTPS configuration."
    else
        error "Run $SCRIPTPATH/$SCRIPT in interactive mode or"
        error "with --forcehttps flag to continue HTTPS configuration."
        exit 1
    fi
fi

debug "Generating self-signed certificates"
mkdir -p $SSL_DIR
CERT_GENERATE=$(openssl req -x509 -subj "/C=US/ST=CA/L=Alameda/O=Perforce/OU=GCONN/CN=$GCONN_HOSTNAME" -nodes -days 365 -newkey rsa:2048 -keyout $SSL_DIR/apache.key -out $SSL_DIR/apache.crt 2>&1)
if [ $? -ne 0 ] && [ -e "$SSL_DIR/apache.key" ] && [ -e "$SSL_DIR/apache.crt" ]; then
    error 'Errors while creating self-signed certificates!'
    echo 'Here is the detailed output:'
    echo "$CERT_GENERATE"
    exit 1
else
    debug 'Self-signed certificates have been created'
fi

APACHE_START=$(service $APACHE_SERVICE_NAME start 2>&1)
if [ $? -ne 0 ]; then
    error "Apache could not start correctly! 'service $APACHE_SERVICE_NAME start'"
    echo 'Here is the detailed output:'
    echo "$APACHE_START"
    exit 1
else
    highlight 'Apache started successfully'
fi
