#!/usr/bin/env python3

import json
import urllib
try:
    import urllib2
    urlparse = urllib
except ImportError:
    import urllib.request as urllib2
    import urllib.parse as urlparse
try:
    import httplib
except ImportError:
    import http.client as httplib
import ssl
import subprocess
import sys
import os
import optparse
import datetime

MIRROR_MSG = """
  "GConn Hook Error: You need to configure:
        'git config gconn.mirror-url  <https://gconnhost/mirror>'
   for the Gerrit repo with GIT_DIR=%s
"""

TOKEN_MSG = """
  "GConn Hook Error: You need to configure:
        'git config gconn.token  <secret-token from GConn mirroring configuration>'
   for the Gerrit repo with GIT_DIR=%s
"""

CONFIG_MSG = """
   GConn Hook Error: You must configure at least one of the two following keys (GConn checks in the order below):
        'git config gconn.git-ssh-url  <ssh  url for this Gerrit repo>'
        'git config gconn.git-http-url <http url for this Gerrit repo>'
   for the Gerrit repo with GIT_DIR=%s
"""

USAGE = """
Description:
    This script issues a Gerrit webhook request to the Helix GConn
    repo mirror url which triggers a 'git fetch' back to the repo's
    Gerrit pull url in order to update the Helix mirror for this repo.
    Helix GConn must be configured for mirroring for this Gerrit repo.

Installation:
    Copy this file to the Gerrit "'$site_path'/hooks" and rename it to 'change-merged'.
    'python' must be in the Gerrit user PATH.

Configuration:
   Enable this webhook for a specific Gerrit repo as follows:
        'cd $site_path/git/<reponame>'
        'git config gconn.mirror-url  <https://gconnhost/mirror>'
        'git config gconn.token <secret-token from GConn mirroring configuration>'
   Configure at least one of the two following keys (GConn checks in the order below):
        'git config gconn.git-ssh-url  <ssh  url for this Gerrit repo>'
        'git config gconn.git-http-url <http url for this Gerrit repo>'

   Disabling http certifcate verification is supported for Python2.7.9 and greater with:
        'git config gconn.http.sslverify false'
   For Python < Python2.7.9 a message will occur in the $site_path/logs/error_log.


   Unconfigured repos will cause no actions.

Command line usage:
    By adding 'GIT_DIR=$site_path/git/reponame' to the environment, one may execute
    this script from the command line for testing, using the logging options as needed.
    No Gerrit options are required to execute from the command line.

Gerrit invocation:
    This hook relies solely on the git repo configuration values describe above in 'Configuration'.
    Gerrit calls the change-merged hook as below. However, all arguments are ignored by this webhook.
    change-merged --change <change id> --change-url <change url> --change-owner <change owner> \\
                  --project <project name> --branch <branch> --topic <topic> --submitter <submitter> --commit <sha1> --newrev <sha1>

Trouble shooting:
   Run the script from the command line after setting GIT_DIR (set above).
   Examine the Gerrit $site_dir/logs/error_log for hook references.
   Ensure the 'hooks' plugin is installed and enabled (see Plugins -> Installed on the Gerrit web interface).
   Installing Gerrit with '--batch' (java -jar gerrit.war init --batch) does not install the Gerrit 'hooks' plugin.
"""

# Logging when running from command line.
# These options cannot be passed by Gerrit.
FILE=None
LOGFILE = "/tmp/gconn-change-merged.log"

def logger(msg, _print=False):
   "Conditionally log/print message."

   if FILE:   # set by --log or --stdout
       date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
       if len(msg):
           msg = "%s : %s" % (date, msg)
       if not msg.endswith('\n'):
           msg = msg + '\n'
       FILE.write(msg)
   elif _print:   # guarantee we get this message back to Gerrit logging
       print(msg)



def gconn_mirror(log, stdout):
    """Issue http request to the GCONN mirroring url for this git repo.
    Configuration is found in the repo's .git/config file. See USAGE.
    """

    global FILE
    if log:
        try:
            FILE = open(LOGFILE, 'a')
        except IOError as e:
            print("Cannot open %s %s " % (logfile, str(e)))
            print("continuing without logging ...")
            pass
    if stdout:
        FILE = sys.stdout

    # Gerrit sets GIT_DIR in the ENV for this repo
    if 'GIT_DIR' in os.environ:
        git_dir = os.environ["GIT_DIR"]
        git_dir_arg = "--git-dir=%s" % (git_dir,)
        logger('')
        logger("GIT_DIR=%s" % git_dir_arg)
    else:
        logger("GConn Hook Error: No GIT_DIR in environment - exit", True)
        return

    # Use the GIT_DIR to get the configured urls from 'git config'.
    git_http_url = ''
    git_ssh_url = ''
    gconn_mirror_url = ''
    gconn_token = ''
    gconn_http_sslverify=''
    try:
        gconn_mirror_url = subprocess.check_output(["git", git_dir_arg, "config","gconn.mirror-url"]).rstrip()
        logger("gconn.mirror-url=%s" % gconn_mirror_url)
    except subprocess.CalledProcessError:
        pass
    if gconn_mirror_url == '':
       msg = MIRROR_MSG % (git_dir,)
       logger(msg, True)
       return

    try:
        gconn_token = subprocess.check_output(["git", git_dir_arg, "config","gconn.token"]).rstrip()
        logger("gconn.token=%s" % gconn_token)
    except subprocess.CalledProcessError:
        pass
    if gconn_token == '':
       msg = TOKEN_MSG % (git_dir,)
       logger(msg, True)
       return

    try:
        git_http_url = subprocess.check_output(["git", git_dir_arg, "config","gconn.git-http-url"]).rstrip()
    except subprocess.CalledProcessError:
        pass
    logger("gconn.git-http-url=%s" % git_http_url)

    try:
        git_ssh_url = subprocess.check_output(["git", git_dir_arg, "config","gconn.git-ssh-url"]).rstrip()
    except subprocess.CalledProcessError:
        pass
    logger("gconn.git-ssh-url=%s" % git_ssh_url)

    # Require at least one Gerrit fetch url.
    if not git_http_url and not git_ssh_url:
       msg = CONFIG_MSG % (git_dir,)
       logger(msg, True)
       return

    try:
        gconn_http_sslverify = subprocess.check_output(["git", git_dir_arg, "config","gconn.http.sslverify"]).rstrip()
        logger("gconn.http.sslverify=%s" % gconn_http_sslverify)
    except subprocess.CalledProcessError:
        pass
    gconn_http_sslverify = False if gconn_http_sslverify.lower() == 'false' else True
    if gconn_http_sslverify and sys.version_info < (2, 7, 9):
        msg = "GConn Hook Error: Setting gconn.http.sslveriy=false requires Python 2.7.9 or greater.\nYou have %s" % (sys.version_info)
        logger(msg, True)
        return

    # Build and issue the http request.
    # Formulate the request to be a subset of the Gitlab webhook request.
    # Doing so utilizes GConn's support for Gitlab webhooks.

    # Set only those key:values required by GConn.
    data = {
        "project":{
            "git_ssh_url" : git_ssh_url,
            "git_http_url": git_http_url
          },
    }

    dataJsonStr = json.dumps(data)
    encoded_data = urlparse.urlencode({"payload": json.dumps(data)})

    # Set the Gitlab event and token headers so GConn understand this as a Gitlab request.
    headers = {'X-Gitlab-Event' : 'Push Hook', 'X-Gitlab-Token': gconn_token, 'Content-Type' : 'application/json'}
    logger("http headers: %s" % (headers,))
    logger("http body:    %s" % (data,))
    request = urllib2.Request(gconn_mirror_url, dataJsonStr.encode(), headers)

    try:
        if gconn_http_sslverify:
            response_obj =  urllib2.urlopen(request)
        else:
            response_obj =  urllib2.urlopen(request, context=ssl._create_unverified_context())
    except urllib2.HTTPError as e:
        logger('GConn Hook HTTPError: = ' + str(e),True)
        return
    except urllib2.URLError as e:
        logger('GConn Hook URLError: ' + str(e), True)
        return
    except httplib.HTTPException as e:
        logger('GConn Hook HTTPException: ' + str(e), True)
        return
    except Exception as e:
        logger('GConn Hook Exception: ' + str(e), True)
        return

    response = response_obj.read()
    logger("GConn Hook HTTP response: " + str(response.rstrip()), True)
    return


def main():
    # Override formatter to retain newlines in usage
    optparse.OptionParser.format_epilog = lambda self, formatter: self.epilog

    parser = optparse.OptionParser(epilog=USAGE)
    parser.add_option('--log', action="store_true", default=False, help="Log to /tmp/change-merge.log")
    parser.add_option('--stdout', action="store_true", default=False, help="Log to stdout")
    parser.add_option("--change",       help="change id    (unused by this GConn hook)")
    parser.add_option("--change-url",   help="change url   (unused by this GConn hook)")
    parser.add_option("--change-owner", help="change owner (unused by this GConn hook)")
    parser.add_option("--project",      help="project name (unused by this GConn hook)")
    parser.add_option("--branch",       help="branch       (unused by this GConn hook)")
    parser.add_option("--topic",        help="topic        (unused by this GConn hook)")
    parser.add_option("--submitter",    help="submitter    (unused by this GConn hook)")
    parser.add_option("--uploader",    help="uploader    (unused by this GConn hook)")
    parser.add_option("--refname",    help="refname    (unused by this GConn hook)")
    parser.add_option("--oldrev",    help="oldrev    (unused by this GConn hook)")
    parser.add_option("--commit",       help="sha1         (unused by this GConn hook)")
    parser.add_option("--newrev",       help="sha1         (unused by this GConn hook)")
    options, remainder = parser.parse_args()
    if options.log and options.stdout:
        print("--stdout and --log may not both be used")
        return
    gconn_mirror(options.log, options.stdout)

if __name__ == "__main__":
    main()
